{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE DuplicateRecordFields #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE QuasiQuotes #-}
{-# LANGUAGE RecordWildCards #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE TypeFamilies #-}

{-# LANGUAGE NoImplicitPrelude #-}

module Datarekisteri.Frontend.Handlers.Applications where

import Relude hiding (id)

import Data.Maybe (fromJust)
import Data.Time (Day)
import Data.Morpheus.Client (raw, declareLocalTypesInline)

import Yesod hiding (emailField)
import Yesod.Auth

import Datarekisteri.Core.Types hiding (Applications)
import Datarekisteri.Frontend.ApiRequests
import Datarekisteri.Frontend.FormFields
import Datarekisteri.Frontend.Types
import Datarekisteri.Frontend.Widgets

declareLocalTypesInline "schema.gql" [raw|
query Applications {
  applications {
    id
    name
    nickname
    email
    phoneNumber
    homeplace
    birthdate
    application
  }
}
|]

declareLocalTypesInline "schema.gql" [raw|
mutation Reject($user: UserID!) {
  reject(user: $user)
}
|]

declareLocalTypesInline "schema.gql" [raw|
mutation Accept($user: UserID!) {
  accept(user: $user)
}
|]

resolveApplicationForm :: Maybe UserID -> Form UserID
resolveApplicationForm user = renderDivs $ areq hiddenField "" user

applicationsW :: [ApplicationsApplications] -> Widget
applicationsW applications = do
    setTitle "Jäsenhakemukset"
    [whamlet|<h1>Käsiteltävät jäsenhakemukset|]
    when (null applications) $ [whamlet|<p>Ei jäsenhakemuksia.|]
    rejectRoute <- handlerToWidget $ maybeAuthorized RejectR True
    acceptRoute <- handlerToWidget $ maybeAuthorized AcceptR True
    forM_ applications $ \ApplicationsApplications {..} -> do
        (widget, enctype) <- handlerToWidget $ generateFormPost $ resolveApplicationForm $ Just id
        let keysAndValues =
                [ ("Nimi" :: Text, name)
                , ("Kutsumanimi", nickname)
                , ("Kotipaikka", homeplace)
                , ("Syntymäaika", renderDate birthdate)
                , ("Sähköposti", renderEmail $ fromJust email)
                , ("Puhelinnumero", renderPhoneNumber phoneNumber)
                ]
            acceptFormContent = [whamlet|
                ^{widget}
                ^{submitButton "Hyväksy"}
                |]
            rejectFormContent = [whamlet|
                ^{widget}
                ^{submitButtonBad "Hylkää"}
                |]
        detailsClass <- newIdent
        [whamlet|
            <details .#{detailsClass}>
                <summary>
                    <h2>#{name}
                <article>
                    <h3>Hakemus
                    <p>#{Textarea application}
                    ^{keyValueTable keysAndValues}
                $maybe route <- acceptRoute
                    ^{form route enctype acceptFormContent}
                $maybe route <- rejectRoute
                    ^{form route enctype rejectFormContent}
        |]
        toWidget [lucius|
            .#{detailsClass} {
                padding: 1.3em;
                margin: 0.6em;
                border-color: #00838a;
                border-width: 0 0 0 0.6em;
                border-style: solid;
                border-radius: 0.4em;
                background-color: #f6f8f8;
                > summary > h2 {
                    display: inline-block;
                    margin-top: 0;
                    margin-bottom: 0;
                }
                > article { margin-bottom: 1.5em }
            }
            .#{detailsClass}[open] { border-color: #339ca1; }
        |]

getApplicationsR :: (YesodAuth DataIdClient, AuthId DataIdClient ~ Text) => Handler Html
getApplicationsR = do
    Applications applications <- apiRequest @Applications True ()
    defaultLayout $ applicationsW applications

postAcceptR :: (YesodAuth DataIdClient, AuthId DataIdClient ~ Text) => Handler Html
postAcceptR = do
    ((result, _), _) <- runFormPost $ resolveApplicationForm Nothing
    case result of
      FormSuccess user -> void $ apiRequest @Accept True $ AcceptArgs user
      _ -> pure ()
    redirect ApplicationsR

postRejectR :: (YesodAuth DataIdClient, AuthId DataIdClient ~ Text) => Handler Html
postRejectR = do
    ((result, _), _) <- runFormPost $ resolveApplicationForm Nothing
    case result of
      FormSuccess user -> void $ apiRequest @Reject True $ RejectArgs user
      _ -> pure ()
    redirect ApplicationsR
