{-# LANGUAGE DeriveAnyClass #-}
{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE QuasiQuotes #-}
{-# LANGUAGE RankNTypes #-}
{-# LANGUAGE RecordWildCards #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE ViewPatterns #-}

{-# LANGUAGE NoImplicitPrelude #-}

module Datarekisteri.Frontend.Types where

import Relude hiding (id)

import Data.Map (findWithDefault)
import Data.Morpheus.Client (raw, declareLocalTypesInline)
import Data.Time (getCurrentTime)
import Data.Time.Format.ISO8601 (iso8601Show)
import Relude.Extra.Foldable1 (maximum1)
import Yesod.Core.Handler (getCurrentRoute)
import Yesod.Default.Util (addStaticContentExternal)

import Yesod
import Yesod.Auth
import Yesod.Static

import Datarekisteri.Core.Types (UserID(..), Scope(..), Permission(..), Permissions(..))
import Datarekisteri.Frontend.ApiRequests
import Datarekisteri.Frontend.Auth

import qualified Datarekisteri.Core.Types as Core

data DataIdClient = DataIdClient
    { getStatic :: Static
    , getConfig :: Config
    }

data Config = Config
    { configPort :: Int
    , configApproot :: Text
    , configServerUrl :: Text
    }

instance PathPiece UserID where
    toPathPiece (UserID id) = show id
    fromPathPiece s = UserID <$> readMaybe (toString s)

mkYesodData "DataIdClient" [parseRoutes|
/ HomeR GET

/profile OwnProfileR GET
/profile/#UserID ProfileR GET POST
/update-password/#UserID UpdatePasswordR POST
/verify-email VerifyEmailR GET POST

/members MembersR GET

/applications ApplicationsR GET
/accept AcceptR POST
/reject RejectR POST

/apply ApplyR GET POST

/auth AuthR Auth getAuth
/static StaticR Static getStatic
|]

declareLocalTypesInline "schema.gql" [raw|
query GetPermissions {
  permissions {
    scope
    permission
  }
  user {
    id
  }
}
|]

declareLocalTypesInline "schema.gql" [raw|
mutation GetWebUIToken {
  newToken(comment: "id.datat.fi webui") {
    tokenData
  }
}
|]

instance YesodAuth DataIdClient where
    type AuthId DataIdClient = Text
    maybeAuthId = lookupSession credsKey
    loginDest = const HomeR
    logoutDest = const HomeR
    authPlugins = const $ [authExternalBasic getToken]
        where getToken auth = (>>= fmap (tokenData . newToken) . rightToMaybe) <$>
                      apiRequestAuth @GetWebUIToken [] (Just auth) ()
    authenticate = pure . Authenticated . credsIdent

withAuthenticated :: (AuthId DataIdClient -> Handler AuthResult) -> Handler AuthResult
withAuthenticated m = maybeAuthId >>= maybe (pure AuthenticationRequired) m

readPermissions :: [GetPermissionsPermissions] -> Permissions
readPermissions = Permissions . fromList . map (\GetPermissionsPermissions {..} -> (scope, permission))

hasPermission :: Scope -> Permission -> Handler Bool
hasPermission scope permission = do
    GetPermissions {..} <- apiRequest @GetPermissions False ()
    let permissions' = readPermissions permissions
        userID = (\x -> id (x :: GetPermissionsUser)) <$> user
        scopes = scope :| case scope of
            Tokens uid | Just uid == userID -> [OwnTokens]
            Profile uid | Just uid == userID -> [OwnProfile]
            _ -> []
            -- TODO add Members and Applications to Profile Scopes if profile owner is member/applicant
    pure $ any (\scope -> Core.hasPermission scope permission permissions') scopes

requirePermission :: Scope -> Permission -> Handler AuthResult
requirePermission scope permission = ifM (hasPermission scope permission)
    (pure Authorized)
    (pure $ Unauthorized $ "Toiminto vaatii " <> renderPermission permission <> " " <> renderScope scope)
        where renderPermission ReadWrite = "kirjoitusoikeuden"
              renderPermission ReadOnly = "lukuoikeuden"
              renderPermission None = "nollaoikeuden"
              renderScope OwnProfile = "omaan profiliin"
              renderScope (Profile _) = "käyttäjän profiliin"
              renderScope OwnTokens = "omiin tokeneihin" -- TODO kunnon suomennos
              renderScope (Tokens _) = "käyttäjän tokeneihin" -- TODO kunnon suomennos
              renderScope Members = "kaikkiin jäseniin"
              renderScope Applications = "jäsenhakemuksiin"

authorizedHelper :: Scope -> Bool -> Handler AuthResult
authorizedHelper scope True = requirePermission scope ReadWrite
authorizedHelper scope False = requirePermission scope ReadOnly

instance Yesod DataIdClient where
    authRoute = const $ Just $ AuthR LoginR
    approot = ApprootMaster $ configApproot . getConfig
    makeSessionBackend = const $ sslOnlySessions $ laxSameSiteSessions $
        Just <$> defaultClientSessionBackend (60*24*7) "client_session_key.aes"
    isAuthorized (AuthR _) _ = pure Authorized
    isAuthorized (StaticR _) _ = pure Authorized
    isAuthorized HomeR _ = pure Authorized
    isAuthorized ApplyR _ = pure Authorized
    isAuthorized VerifyEmailR _ = pure Authorized
    isAuthorized OwnProfileR isWrite = withAuthenticated $ const $ authorizedHelper OwnProfile isWrite
    isAuthorized (ProfileR user) isWrite = withAuthenticated $ const $ authorizedHelper (Profile user) isWrite
    isAuthorized (UpdatePasswordR user) isWrite =
        withAuthenticated $ const $ authorizedHelper (Profile user) isWrite
    isAuthorized MembersR isWrite = withAuthenticated $ const $ authorizedHelper Members isWrite
    isAuthorized ApplicationsR isWrite = withAuthenticated $ const $ authorizedHelper Applications isWrite
    isAuthorized AcceptR isWrite = withAuthenticated $ const $ authorizedHelper Applications isWrite
    isAuthorized RejectR isWrite = withAuthenticated $ const $ authorizedHelper Applications isWrite
    errorHandler NotFound = fmap toTypedContent $ defaultLayout $ defaultMessageWidget "Sivua ei löytynyt" $
        [hamlet|<p>Pyytämästäsi osoitteesta ei löytynyt sivua|]
    errorHandler (InternalError e) = do
        $logErrorS "yesod-core" e
        selectRep $ do
            provideRep $ defaultLayout $ do
                currentTime <- liftIO getCurrentTime
                defaultMessageWidget "Palvelinvirhe" [hamlet|
<p>
    Pyynnön käsittelyssä tapahtui virhe. #
    Ole yhteydessä ylläpitäjiin ja kerro heille palvelimen kellonaika virheen tapahtuessa #
    #{iso8601Show currentTime}
|]
    errorHandler x = defaultErrorHandler x
    addStaticContent = addStaticContentExternal Right base64md5 "/tmp/data-id" (StaticR . flip StaticRoute [])
    defaultLayout widget = do
        let messages :: Widget
            messages = do
                msgs <- getMessages
                messagesClass <- newIdent
                messageClass <- newIdent
                toWidget [hamlet|
                    $if not $ null msgs
                        <aside class="#{messagesClass}">
                          <ul>
                            $forall (_, msg) <- msgs
                              <li class="#{messageClass}">#{msg}
                |]
                toWidget [lucius|
                    aside.#{messagesClass} > ul {
                      list-style: none;
                      padding: 0;
                      margin: 0;
                      li.#{messageClass} {
                        display: block;
                        padding: 0.7em;
                        background-color: #3b4553;
                        color: white;
                        border-radius: 0.3em;
                      }
                    }
                |]
            navigationBar :: Widget
            navigationBar = do
                applicationsRoute <- handlerToWidget $ maybeAuthorized ApplicationsR False
                profileRoute <- handlerToWidget $ maybeAuthorized OwnProfileR False
                membersRoute <- handlerToWidget $ maybeAuthorized MembersR False
                currentRoute <- getCurrentRoute
                loggedIn <- isJust <$> maybeAuthId
                navClass <- newIdent
                activeClass <- newIdent
                rightClass <- newIdent
                let navItem :: Bool -> Maybe (Route DataIdClient) -> Text -> HtmlUrl (Route DataIdClient)
                    navItem alignRight maybeRoute name = [hamlet|
                        $maybe route <- maybeRoute
                            <li :Just route == currentRoute:class="#{activeClass}" :alignRight:class="#{rightClass}">
                                <a href="@{route}">#{name}
                        |]
                    rightNavItem = navItem True
                    leftNavItem = navItem False
                toWidget [hamlet|
                    <nav .#{navClass}>
                      <ul>
                        ^{leftNavItem profileRoute "Profiili"}
                        ^{leftNavItem membersRoute "Jäsenet"}
                        ^{leftNavItem applicationsRoute "Hakemukset"}
                        $if loggedIn
                          ^{rightNavItem (Just (AuthR LogoutR)) "Kirjaudu ulos"}
                        $else
                          ^{rightNavItem (Just (AuthR LoginR)) "Kirjaudu sisään"}
                          ^{rightNavItem (Just ApplyR) "Hae jäseneksi"}
                |]
                toWidget [lucius|
                    .#{navClass} {
                      display: block;
                      position: fixed;
                      top: 0;
                      width: 100%;
                      padding: 0;
                      background-color: #{fgColor};
                      ul {
                        list-style-type: none;
                        margin: 0;
                        padding: 0;
                      }
                      li {
                        display: block;
                        float: left;
                        padding: 0;
                        margin: 0;
                        a {
                          color: #{bgColor};
                          text-decoration: none;
                          display: inline-block;
                          padding: 0.7em 1em;
                        }
                        a:hover {
                          background-color: #00838a;
                          color: #ffffff;
                        }
                      }
                      .#{activeClass} { font-weight: 500; }
                      .#{rightClass} { float: right; }
                    }
                |]

        p <- widgetToPageContent $ do
            [whamlet|
                <header>
                  ^{navigationBar}
                <main>
                  ^{messages}
                  ^{widget}
            |] :: Widget
            toWidget $ [lucius|
                * {
                  box-sizing: border-box;
                }
                html {
                  font-family: "Fira Sans", sans-serif;
                  height: 100%;
                }
                body {
                  background-color: #e8eaef;
                  color: #{fgColor};
                  margin: 0;
                  height: 100%;
                }
                main {
                  margin: 0 auto;
                  max-width: 50em;
                  padding: 5em 1em 3em 1em;
                  background-color: #{bgColor};
                  min-height: 100%;
                  box-sizing: border-box;
                }
                |]
        withUrlRenderer [hamlet|
            $doctype 5
            <html>
              <head>
                <meta name="viewport" content="width=device-width,initial-scale=1"/>
                <title>#{pageTitle p}
                ^{pageHead p}
              <body>
                ^{pageBody p}
            |]

bgColor, fgColor :: Text
bgColor = "#ffffff"
fgColor = "#181c22"

instance ApiRequest DataIdClient where
    getApiUrl = configServerUrl . getConfig <$> getYesod
    authIdToAuthorization = flip const

instance RenderMessage DataIdClient FormMessage where
    renderMessage _ _ = defaultFormMessage

type Form a = Html -> MForm Handler (FormResult a, Widget)
